# Copyright (C) 2020  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from flask import Flask, abort, request

from swh.scanner.exceptions import LargePayloadExc
from swh.scanner.policy import QUERY_LIMIT

from .data import fake_origin, unknown_swhids


def create_app(tmp_requests, tmp_accesses):
    app = Flask(__name__)

    @app.route("/")
    def index():
        return "SWH scanner API"

    @app.route("/known/", methods=["POST"])
    def known():
        swhids = request.get_json()
        with open(tmp_requests, "a") as f:
            for swhid in swhids:
                f.write(swhid + "\n")

        with open(tmp_accesses, "a") as f:
            f.write(f"{len(swhids)}\n")

        if len(swhids) > QUERY_LIMIT:
            raise LargePayloadExc(
                f"The maximum number of SWHIDs this endpoint can receive is "
                f"{QUERY_LIMIT}"
            )

        res = {swhid: {"known": False} for swhid in swhids}
        for swhid in swhids:
            if swhid not in unknown_swhids:
                res[swhid]["known"] = True

        return res

    @app.route("/graph/randomwalk/<swhid>/ori/", methods=["GET"])
    def randomwalk(swhid):
        if swhid in fake_origin.keys():
            return fake_origin[swhid]
        else:
            abort(404)

    return app



# Copyright (C) 2021 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from enum import Enum
import json
import os
import sys
from typing import Any

import ndjson

from swh.model.from_disk import Directory

from .dashboard.dashboard import run_app
from .data import MerkleNodeInfo, get_directory_data
from .plot import generate_sunburst, offline_plot

DEFAULT_OUTPUT = "text"


class Color(Enum):
    BLUE = "\033[94m"
    GREEN = "\033[92m"
    RED = "\033[91m"
    END = "\033[0m"


def colorize(text: str, color: Color):
    return color.value + text + Color.END.value


class Output:
    def __init__(
        self, root_path: str, nodes_data: MerkleNodeInfo, source_tree: Directory
    ):
        self.root_path = root_path
        self.nodes_data = nodes_data
        self.source_tree = source_tree

    def show(self, mode=DEFAULT_OUTPUT):
        if mode == "summary":
            self.summary()
        elif mode == "text":
            isatty = sys.stdout.isatty()
            self.print_text(isatty)
        elif mode == "sunburst":
            directory_data = get_directory_data(
                self.root_path, self.source_tree, self.nodes_data
            )
            sunburst_figure = generate_sunburst(directory_data, self.root_path)
            offline_plot(sunburst_figure)
        elif mode == "interactive":
            directory_data = get_directory_data(
                self.root_path, self.source_tree, self.nodes_data
            )
            sunburst_figure = generate_sunburst(directory_data, self.root_path)
            run_app(sunburst_figure, self.source_tree, self.nodes_data)
        elif mode == "json":
            self.print_json()
        elif mode == "ndjson":
            self.print_ndjson()
        else:
            raise Exception(f"mode {mode} is not an output format")

    def get_path_name(self, node):
        return "path" if "path" in node.data.keys() else "data"

    def print_text(self, isatty: bool) -> None:
        def compute_level(node):
            node_path = str(node.data[self.get_path_name(node)]).split("/")
            source_path = str(self.source_tree.data["path"]).split("/")
            return len(node_path) - len(source_path)

        for node in self.source_tree.iter_tree():
            self.print_node(node, isatty, compute_level(node))

    def print_node(self, node: Any, isatty: bool, level: int) -> None:
        rel_path = os.path.basename(node.data[self.get_path_name(node)])
        rel_path = rel_path.decode()
        begin = "│   " * level
        end = "/" if node.object_type == "directory" else ""

        if isatty:
            if not self.nodes_data[node.swhid()]["known"]:
                rel_path = colorize(rel_path, Color.RED)
            elif node.object_type == "directory":
                rel_path = colorize(rel_path, Color.BLUE)
            elif node.object_type == "content":
                rel_path = colorize(rel_path, Color.GREEN)

        print(f"{begin}{rel_path}{end}")

    def summary(self):
        directories_with_known_files = set()

        total_files = 0
        total_directories = 0
        known_files = 0
        full_known_directories = 0
        partially_known_directories = 0

        contents = []
        directories = []

        for node in self.source_tree.iter_tree():
            if node.object_type == "content":
                contents.append(node)
            elif node.object_type == "directory":
                directories.append(node)
            else:
                assert False, "unreachable"

        total_files = len(contents)
        for c in contents:
            if self.nodes_data[c.swhid()]["known"]:
                known_files += 1
                path = c.data[self.get_path_name(c)]
                dir_name = os.path.dirname(path)
                directories_with_known_files.add(dir_name)

        total_directories = len(directories)
        for d in directories:
            if self.nodes_data[d.swhid()]["known"]:
                full_known_directories += 1
            else:
                path = d.data[self.get_path_name(d)]
                if path in directories_with_known_files:
                    partially_known_directories += 1

        kp = known_files * 100 // total_files
        fkp = full_known_directories * 100 // total_directories
        pkp = partially_known_directories * 100 // total_directories
        print(f"Files:             {total_files:10d}")
        print(f"            known: {known_files:10d} ({kp:3d}%)")
        print(f"directories:       {total_directories:10d}")
        print(f"      fully-known: {full_known_directories:10d} ({fkp:3d}%)")
        print(f"  partially-known: {partially_known_directories:10d} ({pkp:3d}%)")
        print("(see other --output-format for more details)")

    def data_as_json(self):
        json = {}
        for node in self.source_tree.iter_tree():
            rel_path = os.path.relpath(
                node.data[self.get_path_name(node)].decode(),
                self.source_tree.data["path"].decode(),
            )
            json[rel_path] = {"swhid": str(node.swhid())}
            for k, v in self.nodes_data[node.swhid()].items():
                json[rel_path][k] = v
        return json

    def print_json(self):
        print(json.dumps(self.data_as_json(), indent=4, sort_keys=True))

    def print_ndjson(self):
        print(ndjson.dumps({k: v} for k, v in self.data_as_json().items()), flush=True)



# Copyright (C) 2020  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from dash import html

from swh.model.swhids import CoreSWHID, ObjectType
from swh.scanner.dashboard.dashboard import generate_table_body
from swh.scanner.data import MerkleNodeInfo


def test_generate_table_body(source_tree):
    chart_path = b"/bar/barfoo"
    dir_path = source_tree[b"/bar/barfoo"].data["path"].decode()
    nodes_data = MerkleNodeInfo()
    # CoreSWHID of 'another-quote.org'
    known_cnt_swhid = CoreSWHID(
        object_type=ObjectType.CONTENT,
        object_id=b"\x136\x93\xb1%\xba\xd2\xb4\xac1\x855\xb8I\x01\xeb\xb1\xf6\xb68",
    )
    nodes_data[known_cnt_swhid] = {"known": True}

    generated_body = generate_table_body(chart_path, source_tree, nodes_data)

    expected_body = [
        html.Tbody(
            [
                html.Tr(
                    [
                        html.Td("✔"),
                        html.Td(
                            html.A(
                                children="another-quote.org",
                                href=f"file://{dir_path}/another-quote.org",
                            )
                        ),
                        html.Td("swh:1:cnt:133693b125bad2b4ac318535b84901ebb1f6b638"),
                    ]
                ),
            ]
        )
    ]

    # workaround: dash_html_component.__eq__ checks for object identity only
    assert str(generated_body) == str(expected_body)



# Copyright (C) 2021 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import json

import pytest

from swh.model.swhids import CoreSWHID
from swh.scanner.client import Client
from swh.scanner.exceptions import APIError

from .data import correct_known_api_response, correct_origin_api_response

AIO_URL = "http://example.org/api/"
KNOWN_URL = f"{AIO_URL}known/"
ORIGIN_URL = f"{AIO_URL}graph/randomwalk/"


def test_client_known_correct_api_request(mock_aioresponse, event_loop, aiosession):
    mock_aioresponse.post(
        KNOWN_URL,
        status=200,
        content_type="application/json",
        body=json.dumps(correct_known_api_response),
    )

    client = Client(AIO_URL, aiosession)
    actual_result = event_loop.run_until_complete(client.known([]))

    assert correct_known_api_response == actual_result


def test_client_known_raise_apierror(mock_aioresponse, event_loop, aiosession):
    mock_aioresponse.post(KNOWN_URL, content_type="application/json", status=413)

    client = Client(AIO_URL, aiosession)
    with pytest.raises(APIError):
        event_loop.run_until_complete(client.known([]))


def test_client_get_origin_correct_api_request(
    mock_aioresponse, event_loop, aiosession
):
    origin_url = (
        f"{ORIGIN_URL}swh:1:dir:01fa282bb80be5907505d44b4692d3fa40fad140/ori"
        f"/?direction=backward&limit=-1&resolve_origins=true"
    )
    mock_aioresponse.get(
        origin_url,
        status=200,
        body=correct_origin_api_response,
    )

    client = Client(AIO_URL, aiosession)
    swhid = CoreSWHID.from_string("swh:1:dir:01fa282bb80be5907505d44b4692d3fa40fad140")
    actual_result = event_loop.run_until_complete(client.get_origin(swhid))

    assert correct_origin_api_response == actual_result



# Copyright (C) 2021 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import abc
import itertools
from typing import Iterable, List, no_type_check

from swh.core.utils import grouper
from swh.model import discovery, from_disk
from swh.model.from_disk import model
from swh.model.model import Sha1Git

from .client import QUERY_LIMIT, Client
from .data import MerkleNodeInfo


def source_size(source_tree: from_disk.Directory):
    """return the size of a source tree as the number of nodes it contains"""
    return sum(1 for n in source_tree.iter_tree(dedup=False))


class Policy(metaclass=abc.ABCMeta):

    data: MerkleNodeInfo
    """information about contents and directories of the merkle tree"""

    source_tree: from_disk.Directory
    """representation of a source code project directory in the merkle tree"""

    def __init__(self, source_tree: from_disk.Directory, data: MerkleNodeInfo):
        self.source_tree = source_tree
        self.data = data

    @abc.abstractmethod
    async def run(self, client: Client):
        """Scan a source code project"""
        raise NotImplementedError("Must implement run method")


class LazyBFS(Policy):
    """Read nodes in the merkle tree using the BFS algorithm.
    Lookup only directories that are unknown otherwise set all the downstream
    contents to known.
    """

    async def run(self, client: Client):
        queue = []
        queue.append(self.source_tree)

        while queue:
            swhids = [node.swhid() for node in queue]
            swhids_res = await client.known(swhids)
            for node in queue.copy():
                queue.remove(node)
                self.data[node.swhid()]["known"] = swhids_res[str(node.swhid())][
                    "known"
                ]
                if node.object_type == "directory":
                    if not self.data[node.swhid()]["known"]:
                        children = [n[1] for n in list(node.items())]
                        queue.extend(children)
                    else:
                        for sub_node in node.iter_tree():
                            if sub_node == node:
                                continue
                            self.data[sub_node.swhid()]["known"] = True  # type: ignore


class GreedyBFS(Policy):
    """Query graph nodes in chunks (to maximize the Web API rate limit use) and set the
    downstream contents of known directories to known.
    """

    async def run(self, client: Client):
        ssize = source_size(self.source_tree)
        seen = []

        async for nodes_chunk in self.get_nodes_chunks(client, ssize):
            for node in nodes_chunk:
                seen.append(node)
                if len(seen) == ssize:
                    return
                if node.object_type == "directory" and self.data[node.swhid()]["known"]:
                    sub_nodes = [n for n in node.iter_tree(dedup=False)]
                    sub_nodes.remove(node)  # remove root node
                    for sub_node in sub_nodes:
                        seen.append(sub_node)
                        self.data[sub_node.swhid()]["known"] = True

    @no_type_check
    async def get_nodes_chunks(self, client: Client, ssize: int):
        """Query chunks of QUERY_LIMIT nodes at once in order to fill the Web API
        rate limit. It query all the nodes in the case the source code contains
        less than QUERY_LIMIT nodes.
        """
        nodes = self.source_tree.iter_tree(dedup=False)
        for nodes_chunk in grouper(nodes, QUERY_LIMIT):
            nodes_chunk = [n for n in nodes_chunk]
            swhids = [node.swhid() for node in nodes_chunk]
            swhids_res = await client.known(swhids)
            for node in nodes_chunk:
                swhid = node.swhid()
                self.data[swhid]["known"] = swhids_res[str(swhid)]["known"]
            yield nodes_chunk


class FilePriority(Policy):
    """Check the Merkle tree querying all the file contents and set all the upstream
    directories to unknown in the case a file content is unknown.
    Finally check all the directories which status is still unknown and set all the
    sub-directories of known directories to known.
    """

    @no_type_check
    async def run(self, client: Client):
        # get all the files
        all_contents = list(
            filter(
                lambda node: node.object_type == "content", self.source_tree.iter_tree()
            )
        )
        all_contents.reverse()  # check deepest node first

        # query the backend to get all file contents status
        cnt_swhids = [node.swhid() for node in all_contents]
        cnt_status_res = await client.known(cnt_swhids)
        # set all the file contents status
        for cnt in all_contents:
            self.data[cnt.swhid()]["known"] = cnt_status_res[str(cnt.swhid())]["known"]
            # set all the upstream directories of unknown file contents to unknown
            if not self.data[cnt.swhid()]["known"]:
                parent = cnt.parents[0]
                while parent:
                    self.data[parent.swhid()]["known"] = False
                    parent = parent.parents[0] if parent.parents else None

        # get all unset directories and check their status
        # (update children directories accordingly)
        unset_dirs = list(
            filter(
                lambda node: node.object_type == "directory"
                and self.data[node.swhid()]["known"] is None,
                self.source_tree.iter_tree(),
            )
        )

        # check unset directories
        for dir_ in unset_dirs:
            if self.data[dir_.swhid()]["known"] is None:
                # update directory status
                dir_status = await client.known([dir_.swhid()])
                dir_known = dir_status[str(dir_.swhid())]["known"]
                self.data[dir_.swhid()]["known"] = dir_known
                if dir_known:
                    sub_dirs = list(
                        filter(
                            lambda n: n.object_type == "directory"
                            and self.data[n.swhid()]["known"] is None,
                            dir_.iter_tree(),
                        )
                    )
                    for node in sub_dirs:
                        self.data[node.swhid()]["known"] = True


class DirectoryPriority(Policy):
    """Check the Merkle tree querying all the directories that have at least one file
    content and set all the upstream directories to unknown in the case a directory
    is unknown otherwise set all the downstream contents to known.
    Finally check the status of empty directories and all the remaining file
    contents.
    """

    @no_type_check
    async def run(self, client: Client):
        # get all directory contents that have at least one file content
        unknown_dirs = list(
            filter(
                lambda dir_: dir_.object_type == "directory"
                and self.has_contents(dir_),
                self.source_tree.iter_tree(),
            )
        )
        unknown_dirs.reverse()  # check deepest node first

        for dir_ in unknown_dirs:
            if self.data[dir_.swhid()]["known"] is None:
                dir_status = await client.known([dir_.swhid()])
                dir_known = dir_status[str(dir_.swhid())]["known"]
                self.data[dir_.swhid()]["known"] = dir_known
                # set all the downstream file contents to known
                if dir_known:
                    for cnt in self.get_contents(dir_):
                        self.data[cnt.swhid()]["known"] = True
                # otherwise set all the upstream directories to unknown
                else:
                    parent = dir_.parents[0]
                    while parent:
                        self.data[parent.swhid()]["known"] = False
                        parent = parent.parents[0] if parent.parents else None

        # get remaining directories that have no file contents
        empty_dirs = list(
            filter(
                lambda n: n.object_type == "directory"
                and not self.has_contents(n)
                and self.data[n.swhid()]["known"] is None,
                self.source_tree.iter_tree(),
            )
        )
        empty_dirs_swhids = [n.swhid() for n in empty_dirs]
        empty_dir_status = await client.known(empty_dirs_swhids)

        # update status of directories that have no file contents
        for dir_ in empty_dirs:
            self.data[dir_.swhid()]["known"] = empty_dir_status[str(dir_.swhid())][
                "known"
            ]

        # check unknown file contents
        unknown_cnts = list(
            filter(
                lambda n: n.object_type == "content"
                and self.data[n.swhid()]["known"] is None,
                self.source_tree.iter_tree(),
            )
        )
        unknown_cnts_swhids = [n.swhid() for n in unknown_cnts]
        unknown_cnts_status = await client.known(unknown_cnts_swhids)

        for cnt in unknown_cnts:
            self.data[cnt.swhid()]["known"] = unknown_cnts_status[str(cnt.swhid())][
                "known"
            ]

    def has_contents(self, directory: from_disk.Directory):
        """Check if the directory given in input has contents"""
        for entry in directory.entries:
            if entry["type"] == "file":
                return True
        return False

    def get_contents(self, dir_: from_disk.Directory):
        """Get all the contents of a given directory"""
        for _, node in list(dir_.items()):
            if node.object_type == "content":
                yield node


class WebAPIConnection(discovery.ArchiveDiscoveryInterface):
    """Use the web APIs to query the archive"""

    def __init__(
        self,
        contents: List[model.Content],
        skipped_contents: List[model.SkippedContent],
        directories: List[model.Directory],
        client: Client,
    ) -> None:
        self.contents = contents
        self.skipped_contents = skipped_contents
        self.directories = directories
        self.client = client

        self.sha_to_swhid = {}
        self.swhid_to_sha = {}
        for content in contents:
            swhid = str(content.swhid())
            self.sha_to_swhid[content.sha1_git] = swhid
            self.swhid_to_sha[swhid] = content.sha1_git

        for directory in directories:
            swhid = str(directory.swhid())
            self.sha_to_swhid[directory.id] = swhid
            self.swhid_to_sha[swhid] = directory.id

    async def content_missing(self, contents: List[Sha1Git]) -> List[Sha1Git]:
        """List content missing from the archive by sha1"""
        return await self._missing(contents)

    async def skipped_content_missing(
        self, skipped_contents: List[Sha1Git]
    ) -> Iterable[Sha1Git]:
        """List skipped content missing from the archive by sha1"""
        # TODO what should we do about skipped contents?
        return skipped_contents

    async def directory_missing(self, directories: List[Sha1Git]) -> Iterable[Sha1Git]:
        """List directories missing from the archive by sha1"""
        return await self._missing(directories)

    async def _missing(self, shas: List[Sha1Git]) -> List[Sha1Git]:
        # Ignore mypy complaining about string being passed, since `known`
        # transforms them to string immediately.
        res = await self.client.known([self.sha_to_swhid[o] for o in shas])  # type: ignore
        return [self.swhid_to_sha[k] for k, v in res.items() if not v["known"]]


class RandomDirSamplingPriority(Policy):
    """Check the Merkle tree querying random directories. Set all ancestors to
    unknown for unknown directories, otherwise set all descendants to known.
    Finally check all the remaining file contents.
    """

    @no_type_check
    async def run(self, client: Client):
        contents, skipped_contents, directories = from_disk.iter_directory(
            self.source_tree
        )

        # `filter_known_objects` only does filtering by random directory
        # sampling for now.
        # In the future, it could/will grow a parameter to choose/pass in a
        # different discovery implementation.
        # From this call site, we are relying on this behavior in order to
        # *actually* be a random directory sampling policy, but any change away
        # from under us in `filter_known_objects` should trigger a test failure.
        get_unknowns = discovery.filter_known_objects(
            WebAPIConnection(contents, skipped_contents, directories, client),
        )

        unknowns = set(itertools.chain(*await get_unknowns))

        for obj in itertools.chain(contents, skipped_contents, directories):
            self.data[obj.swhid()]["known"] = obj not in unknowns


class QueryAll(Policy):
    """Check the status of every node in the Merkle tree."""

    @no_type_check
    async def run(self, client: Client):
        all_nodes = [node for node in self.source_tree.iter_tree()]
        all_swhids = [node.swhid() for node in all_nodes]
        swhids_res = await client.known(all_swhids)
        for node in all_nodes:
            self.data[node.swhid()]["known"] = swhids_res[str(node.swhid())]["known"]



# Copyright (C) 2021 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

"""
Minimal async web client for the Software Heritage Web API.

This module could be removed when
`T2635 <https://forge.softwareheritage.org/T2635>` is implemented.
"""

import asyncio
import itertools
import logging
import time
from typing import Any, Dict, List, Optional, Tuple

import aiohttp

from swh.model.swhids import CoreSWHID

from .exceptions import error_response

logger = logging.getLogger(__name__)

# Maximum number of SWHIDs that can be requested by a single call to the
# Web API endpoint /known/
QUERY_LIMIT = 1000
MAX_RETRY = 10

KNOWN_EP = "known/"
GRAPH_RANDOMWALK_EP = "graph/randomwalk/"


def _get_chunk(swhids):
    """slice a list of `swhids` into smaller list of size QUERY_LIMIT"""
    for i in range(0, len(swhids), QUERY_LIMIT):
        yield swhids[i : i + QUERY_LIMIT]


def _parse_limit_header(response) -> Tuple[Optional[int], Optional[int], Optional[int]]:
    """parse the X-RateLimit Headers if any"""
    limit = response.headers.get("X-RateLimit-Limit")
    if limit is not None:
        limit = int(limit)
    remaining = response.headers.get("X-RateLimit-Remaining")
    if remaining is not None:
        remaining = int(remaining)
    reset = response.headers.get("X-RateLimit-Reset")
    if reset is not None:
        reset = int(reset)
    return (limit, remaining, reset)


class Client:
    """Manage requests to the Software Heritage Web API."""

    def __init__(
        self,
        api_url: str,
        session: aiohttp.ClientSession,
    ):
        self._sleep = 0
        self.api_url = api_url
        self.session = session
        self._known_endpoint = self.api_url + KNOWN_EP

    async def get_origin(self, swhid: CoreSWHID) -> Optional[Any]:
        """Walk the compressed graph to discover the origin of a given swhid"""
        endpoint = (
            f"{self.api_url}{GRAPH_RANDOMWALK_EP}{str(swhid)}/ori/?direction="
            f"backward&limit=-1&resolve_origins=true"
        )
        res = None
        async with self.session.get(endpoint) as resp:
            if resp.status == 200:
                res = await resp.text()
                res = res.rstrip()
                return res
            if resp.status != 404:
                error_response(resp.reason, resp.status, endpoint)

        return res

    async def known(self, swhids: List[CoreSWHID]) -> Dict[str, Dict[str, bool]]:
        """API Request to get information about the SoftWare Heritage persistent
        IDentifiers (SWHIDs) given in input.

        Args:
            swhids: a list of CoreSWHID instances
            api_url: url for the API request

        Returns:
            A dictionary with:

            key:
                string SWHID searched
            value:
                value['known'] = True if the SWHID is found
                value['known'] = False if the SWHID is not found

        """
        requests = []

        swh_ids = [str(swhid) for swhid in swhids]

        if len(swhids) <= QUERY_LIMIT:
            return await self._make_request(swh_ids)
        else:
            for swhids_chunk in _get_chunk(swh_ids):
                task = asyncio.create_task(self._make_request(swhids_chunk))
                requests.append(task)

            res = await asyncio.gather(*requests)
            # concatenate list of dictionaries
            return dict(itertools.chain.from_iterable(e.items() for e in res))

    def _mark_success(self, limit=None, remaining=None, reset=None):
        """call when a request is successfully made, this will adjust the rate

        The extra argument can be used to transmit the X-RateLimit information
        from the server.  This will be used to adjust the request rate"""
        is_dbg = logger.isEnabledFor(logging.DEBUG)
        self._sleep = 0
        factor = 0
        current = time.time()
        if is_dbg:
            dbg_msg = f"HTTP GOOD {current:.2f}:"
        if limit is None or remaining is None or reset is None:
            if is_dbg:
                dbg_msg += " no rate limit data;"
        else:
            time_windows = reset - current
            if is_dbg:
                dbg_msg += f" requests={remaining}/{limit}"
                dbg_msg += f" reset-in={time_windows:.2f}"
            if time_windows > 0:
                used_up = remaining / limit
                if remaining <= 0:
                    # no more credit, we can sit up and wait.
                    #
                    # XXX we should warn the user. This can get very long.
                    self._sleep = time_windows
                    factor = -1
                elif 0.6 < used_up:
                    # let us not limit the first flight of request.
                    factor = 0
                else:
                    # the deeper we consume the credit the higher is the rate
                    # limiting, let's put a brake on our current rate the lower we get
                    #
                    # (The factor range from 1 to 1000)
                    factor = (0.4 + used_up) ** -1.5
                if factor >= 0:
                    self._sleep = ((time_windows / remaining)) * factor
        if is_dbg:
            dbg_msg += f"; sleep={self._sleep:.3f}"
            logger.debug(dbg_msg)

    def _mark_failure(self, limit=None, remaining=None, reset=None):
        """call when a request failed, this will reduce the request rate.

        The extra argument can be used to transmit the X-RateLimit information
        from the server.  This will be used to adjust the request rate"""
        is_dbg = logger.isEnabledFor(logging.DEBUG)
        current = time.time()
        if is_dbg:
            dbg_msg = f"HTTP BAD  {current:.2f}:"
        time_set = False
        if remaining is None or reset is None:
            if is_dbg:
                dbg_msg += " no rate limit data"
        else:
            wait_for = reset - current
            if is_dbg:
                dbg_msg += f" requests={remaining}/{limit}"
                dbg_msg += f" reset-in={wait_for:.2f}"
            if remaining <= 0:
                # Add some margin to please the rate limiting code
                wait_for *= 1.1
                if wait_for > 0 and wait_for >= self._sleep:
                    self._sleep = wait_for
                    time_set = True
        if not time_set:
            if self._sleep <= 0:
                self._sleep = 1
            else:
                self._sleep *= 2
        if is_dbg:
            dbg_msg += "; sleep={self._sleep:.3f}"
            logger.debug(dbg_msg)

    async def _make_request(self, swhids):
        endpoint = self._known_endpoint

        data = None

        retry = MAX_RETRY

        while data is None:
            # slow the pace of request if needed
            if self._sleep > 0:
                time.sleep(self._sleep)
            async with self.session.post(endpoint, json=swhids) as resp:
                rate_limit = _parse_limit_header(resp)
                if resp.status == 200:
                    try:
                        # inform of success before the await
                        self._mark_success(*rate_limit)
                        data = await resp.json()
                    except aiohttp.client_exceptions.ClientConnectionError:
                        raise
                    else:
                        break
                self._mark_failure(*rate_limit)
                retry -= 1
                if retry <= 0 or resp.status == 413:  # 413: Payload Too Large
                    error_response(resp.reason, resp.status, endpoint)
        return data



# Copyright (C) 2020-2021 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from pathlib import Path

import dash
from dash import dcc, html
from dash.dependencies import Input, Output
import dash_bootstrap_components as dbc
import plotly.graph_objects as go

from swh.model.from_disk import Directory

from ..data import MerkleNodeInfo, get_content_from


def generate_table_body(
    dir_path: bytes, source_tree: Directory, nodes_data: MerkleNodeInfo
):
    """
    Generate the data_table from the path taken from the chart.

    For each file builds the html table rows showing the known status, a local link to
    the file and the relative SoftWare Heritage persistent IDentifier (SWHID).
    """
    contents = get_content_from(dir_path, source_tree, nodes_data)
    data = []
    for cnt, attr in contents.items():
        file_path = Path(cnt.decode())
        file_name = file_path.parts[len(file_path.parts) - 1]
        full_file_path = Path(Path(dir_path.decode()), file_path)
        data.append(
            html.Tr(
                [
                    html.Td("✔" if attr["known"] else ""),
                    html.Td(html.A(file_name, href="file://" + str(full_file_path))),
                    html.Td(attr["swhid"]),
                ]
            )
        )

    return [html.Tbody(data)]


def run_app(graph_obj: go, source_tree: Directory, nodes_data: MerkleNodeInfo):
    app = dash.Dash(__name__)
    fig = go.Figure().add_trace(graph_obj)

    fig.update_layout(
        height=800,
    )

    table_header = [
        html.Thead(html.Tr([html.Th("KNOWN"), html.Th("FILE NAME"), html.Th("SWHID")]))
    ]

    app.layout = html.Div(
        [
            html.Div(
                [
                    html.Div(
                        [
                            dcc.Graph(id="sunburst_chart", figure=fig),
                        ],
                        className="col",
                    ),
                    html.Div(
                        [
                            html.H3(id="directory_title"),
                            dbc.Table(
                                id="files_table",
                                hover=True,
                                responsive=True,
                                striped=True,
                            ),
                        ],
                        className="col",
                    ),
                ],
                className="row",
            ),
        ]
    )

    @app.callback(
        [Output("files_table", "children"), Output("directory_title", "children")],
        [Input("sunburst_chart", "clickData")],
    )
    def update_files_table(click_data):
        """
        Callback that takes the input (directory path) from the chart and
        update the `files_table` children with the relative files.

        """
        if click_data is not None:
            full_path = click_data["points"][0]["label"]
            return (
                table_header
                + generate_table_body(full_path.encode(), source_tree, nodes_data),
                full_path,
            )
        else:
            return "", ""

    app.run_server(debug=True, use_reloader=True)



# Copyright (C) 2021 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from pathlib import Path
from typing import Dict, Optional, Tuple

from swh.model.exceptions import ValidationError
from swh.model.from_disk import Directory
from swh.model.swhids import CoreSWHID

from .client import Client

SUPPORTED_INFO = {"known", "origin"}


class MerkleNodeInfo(dict):
    """Store additional information about Merkle DAG nodes, using SWHIDs as keys"""

    def __setitem__(self, key, value):
        """The keys must be valid valid Software Heritage Persistent Identifiers
        while values must be dict.
        """
        if not isinstance(key, CoreSWHID):
            raise ValidationError("keys must be valid SWHID(s)")

        if not isinstance(value, dict):
            raise ValidationError(f"values must be dict, not {type(value)}")

        super(MerkleNodeInfo, self).__setitem__(key, value)


def init_merkle_node_info(source_tree: Directory, data: MerkleNodeInfo, info: set):
    """Populate the MerkleNodeInfo with the SWHIDs of the given source tree and the
    attributes that will be stored.
    """
    if not info:
        raise Exception("Data initialization requires node attributes values.")
    nodes_info: Dict[str, Optional[str]] = {}
    for ainfo in info:
        if ainfo in SUPPORTED_INFO:
            nodes_info[ainfo] = None
        else:
            raise Exception(f"Information {ainfo} is not supported.")

    for node in source_tree.iter_tree():
        data[node.swhid()] = nodes_info.copy()  # type: ignore


async def add_origin(source_tree: Directory, data: MerkleNodeInfo, client: Client):
    """Store origin information about software artifacts retrieved from the Software
    Heritage graph service.
    """
    queue = []
    queue.append(source_tree)
    while queue:
        for node in queue.copy():
            queue.remove(node)
            node_ori = await client.get_origin(node.swhid())
            if node_ori:
                data[node.swhid()]["origin"] = node_ori
                if node.object_type == "directory":
                    for sub_node in node.iter_tree():
                        data[sub_node.swhid()]["origin"] = node_ori  # type: ignore
            else:
                if node.object_type == "directory":
                    children = [sub_node for sub_node in node.iter_tree()]
                    children.remove(node)
                    queue.extend(children)  # type: ignore


def get_directory_data(
    root_path: str,
    source_tree: Directory,
    nodes_data: MerkleNodeInfo,
    directory_data: Dict = {},
) -> Dict[Path, dict]:
    """Get content information for each directory inside source_tree.

    Returns:
     A dictionary with a directory path as key and the relative
     contents information as values.
    """

    def _get_directory_data(
        source_tree: Directory, nodes_data: MerkleNodeInfo, directory_data: Dict
    ):
        directories = list(
            filter(
                lambda n: n.object_type == "directory",
                map(lambda n: n[1], source_tree.items()),
            )
        )
        for node in directories:
            directory_info = directory_content(node, nodes_data)
            rel_path = Path(node.data["path"].decode()).relative_to(Path(root_path))
            directory_data[rel_path] = directory_info
            if has_dirs(node):
                _get_directory_data(node, nodes_data, directory_data)

    _get_directory_data(source_tree, nodes_data, directory_data)
    return directory_data


def directory_content(node: Directory, nodes_data: MerkleNodeInfo) -> Tuple[int, int]:
    """Count known contents inside the given directory.

    Returns:
     A tuple with the total number of contents inside the directory and the number
     of known contents.
    """
    known_cnt = 0
    node_contents = list(
        filter(lambda n: n.object_type == "content", map(lambda n: n[1], node.items()))
    )
    for sub_node in node_contents:
        if nodes_data[sub_node.swhid()]["known"]:
            known_cnt += 1

    return (len(node_contents), known_cnt)


def has_dirs(node: Directory) -> bool:
    """Check if the given directory has other directories inside."""
    for _, sub_node in node.items():
        if isinstance(sub_node, Directory):
            return True
    return False


def get_content_from(
    node_path: bytes, source_tree: Directory, nodes_data: MerkleNodeInfo
) -> Dict[bytes, dict]:
    """Get content information from the given directory node."""
    # root in model.from_disk.Directory should be accessed with b""
    directory = source_tree[node_path if node_path != source_tree.data["path"] else b""]
    node_contents = list(
        filter(
            lambda n: n.object_type == "content", map(lambda n: n[1], directory.items())
        )
    )
    files_data = {}
    for node in node_contents:
        node_info = nodes_data[node.swhid()]
        node_info["swhid"] = str(node.swhid())
        path_name = "path" if "path" in node.data.keys() else "data"
        files_data[node.data[path_name]] = node_info

    return files_data



# Copyright (C) 2020-2021  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import asyncio
import os
from pathlib import Path
import shutil
import sys

import aiohttp
from aioresponses import aioresponses
import pytest

from swh.model.cli import model_of_dir
from swh.scanner.data import MerkleNodeInfo
from swh.scanner.policy import QUERY_LIMIT

from .data import present_swhids
from .flask_api import create_app


@pytest.fixture
def mock_aioresponse():
    with aioresponses() as m:
        yield m


@pytest.fixture
def event_loop():
    """Fixture that generate an asyncio event loop."""
    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)
    yield loop
    loop.close()


@pytest.fixture
async def aiosession():
    """Fixture that generate an aiohttp Client Session."""
    session = aiohttp.ClientSession()
    yield session
    session.detach()


@pytest.fixture(scope="function")
def test_sample_folder(datadir, tmp_path):
    """Location of the "data" folder"""
    archive_path = Path(os.path.join(datadir, "sample-folder.tgz"))
    assert archive_path.exists()
    shutil.unpack_archive(archive_path, extract_dir=tmp_path)
    test_sample_folder = Path(os.path.join(tmp_path, "sample-folder"))
    assert test_sample_folder.exists()
    return test_sample_folder


@pytest.fixture(scope="function")
def test_sample_folder_policy(datadir, tmp_path):
    """Location of the sample source code project to test the scanner policies"""
    archive_path = Path(os.path.join(datadir, "sample-folder-policy.tgz"))
    assert archive_path.exists()
    shutil.unpack_archive(archive_path, extract_dir=tmp_path)
    test_sample_folder = Path(os.path.join(tmp_path, "sample-folder-policy"))
    assert test_sample_folder.exists()
    return test_sample_folder


@pytest.fixture(scope="function")
def source_tree(test_sample_folder):
    """Generate a model.from_disk.Directory object from the test sample
    folder
    """
    return model_of_dir(str(test_sample_folder).encode())


@pytest.fixture(scope="function")
def big_source_tree(tmp_path):
    """Generate a model.from_disk.Directory from a "big" temporary directory
    (more than 1000 nodes)
    """
    # workaround to avoid a RecursionError that could be generated while creating
    # a large number of directories
    sys.setrecursionlimit(1100)
    dir_ = tmp_path / "big-directory"
    sub_dirs = dir_
    for i in range(0, QUERY_LIMIT + 1):
        sub_dirs = sub_dirs / "dir"
    sub_dirs.mkdir(parents=True, exist_ok=True)
    file_ = sub_dirs / "file.org"
    file_.touch()
    dir_obj = model_of_dir(str(dir_).encode())
    return dir_obj


@pytest.fixture(scope="function")
def source_tree_policy(test_sample_folder_policy):
    """Generate a model.from_disk.Directory object from the test sample
    folder
    """
    return model_of_dir(str(test_sample_folder_policy).encode())


@pytest.fixture(scope="function")
def source_tree_dirs(source_tree):
    """Returns a list of all directories contained inside the test sample
    folder
    """
    root = source_tree.data["path"]
    return list(
        map(
            lambda n: Path(n.data["path"].decode()).relative_to(Path(root.decode())),
            filter(
                lambda n: n.object_type == "directory"
                and not n.data["path"] == source_tree.data["path"],
                source_tree.iter_tree(dedup=False),
            ),
        )
    )


@pytest.fixture(scope="function")
def nodes_data(source_tree):
    """mock known status of file/dirs in test_sample_folder"""
    nodes_data = MerkleNodeInfo()
    for node in source_tree.iter_tree():
        nodes_data[node.swhid()] = {"known": True}
    return nodes_data


@pytest.fixture
def test_swhids_sample(tmp_path):
    """Create and return the opened "swhids_sample" file,
    filled with present swhids present in data.py
    """
    test_swhids_sample = Path(os.path.join(tmp_path, "swhids_sample.txt"))

    with open(test_swhids_sample, "w") as f:
        f.write("\n".join(swhid for swhid in present_swhids))

    assert test_swhids_sample.exists()
    return open(test_swhids_sample, "r")


@pytest.fixture(scope="session")
def tmp_requests(tmpdir_factory):
    """Logs each SWHID per line in every request made to the `known` endpoint"""
    requests_file = tmpdir_factory.mktemp("data").join("requests.json")
    return requests_file


@pytest.fixture(scope="session")
def tmp_accesses(tmpdir_factory):
    """Logs each request made to the `known` endpoint, writing the number
    of SWHIDs queried, one per line."""
    requests_file = tmpdir_factory.mktemp("data").join("accesses.json")
    return requests_file


@pytest.fixture(scope="session")
def app(tmp_requests, tmp_accesses):
    """Flask backend API (used by live_server)."""
    app = create_app(tmp_requests, tmp_accesses)
    return app



# Copyright (C) 2021 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from pathlib import Path

from flask import url_for
import pytest

from swh.model.exceptions import ValidationError
from swh.scanner.client import Client
from swh.scanner.data import (
    MerkleNodeInfo,
    add_origin,
    directory_content,
    get_directory_data,
    has_dirs,
    init_merkle_node_info,
)

from .data import fake_origin


def test_merkle_node_data_wrong_args():
    nodes_data = MerkleNodeInfo()

    with pytest.raises(ValidationError):
        nodes_data["wrong key"] = {"known": True}

    with pytest.raises(ValidationError):
        nodes_data["swh:1:dir:17d207da3804cc60a77cba58e76c3b2f767cb112"] = "wrong value"


def test_init_merkle_supported_node_info(source_tree):
    nodes_data = MerkleNodeInfo()
    init_merkle_node_info(source_tree, nodes_data, {"known", "origin"})
    for _, node_attrs in nodes_data.items():
        assert "known" and "origin" in node_attrs.keys()


def test_init_merkle_not_supported_node_info(source_tree):
    nodes_data = MerkleNodeInfo()
    with pytest.raises(Exception):
        init_merkle_node_info(source_tree, nodes_data, {"unsupported_info"})


def test_add_origin(event_loop, live_server, aiosession, source_tree, nodes_data):
    api_url = url_for("index", _external=True)
    init_merkle_node_info(source_tree, nodes_data, {"known", "origin"})
    client = Client(api_url, aiosession)

    event_loop.run_until_complete(add_origin(source_tree, nodes_data, client))
    for node, attrs in nodes_data.items():
        assert attrs["origin"] == fake_origin[str(source_tree.swhid())]


def test_get_directory_data(source_tree, nodes_data):
    root = Path(source_tree.data["path"].decode())
    dirs_data = get_directory_data(root, source_tree, nodes_data)

    assert len(dirs_data) == 5


def test_directory_content(source_tree, nodes_data):
    foo_dir = source_tree[b"foo"]
    foo_content = directory_content(foo_dir, nodes_data)
    assert foo_content[0] == 3
    assert foo_content[1] == 3


def test_has_dirs(source_tree):
    assert has_dirs(source_tree)



# Copyright (C) 2021 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information


from flask import url_for
import pytest

from swh.model.swhids import CoreSWHID, ObjectType
from swh.scanner.client import Client
from swh.scanner.data import MerkleNodeInfo, init_merkle_node_info
from swh.scanner.policy import (
    DirectoryPriority,
    FilePriority,
    GreedyBFS,
    LazyBFS,
    RandomDirSamplingPriority,
    source_size,
)


def test_scanner_directory_priority_has_contents(source_tree):
    nodes_data = MerkleNodeInfo()
    policy = DirectoryPriority(source_tree, nodes_data)
    assert policy.has_contents(source_tree[b"/bar/barfoo"])


def get_backend_swhids_order(tmp_requests):
    with open(tmp_requests, "r") as f:
        backend_swhids_order = f.readlines()

    return [x.strip() for x in backend_swhids_order]


def get_backend_known_requests(tmp_accesses):
    with open(tmp_accesses, "r") as f:
        calls = f.readlines()

    return [int(call.strip()) for call in calls]


def test_lazybfs_policy(
    live_server, aiosession, event_loop, source_tree_policy, tmp_requests
):
    open(tmp_requests, "w").close()
    api_url = url_for("index", _external=True)

    nodes_data = MerkleNodeInfo()
    init_merkle_node_info(source_tree_policy, nodes_data, {"known"})
    policy = LazyBFS(source_tree_policy, nodes_data)
    client = Client(api_url, aiosession)
    event_loop.run_until_complete(policy.run(client))

    backend_swhids_requests = get_backend_swhids_order(tmp_requests)

    assert (
        backend_swhids_requests[0]
        == "swh:1:dir:fe8cd7076bef324eb8865f818ef08617879022ce"
    )

    # the second request must contain 3 SWHIDs related to directories and one content
    dir_count, cnt_count = 0, 0
    for swhid in backend_swhids_requests[1:5]:
        if CoreSWHID.from_string(swhid).object_type == ObjectType.DIRECTORY:
            dir_count += 1
        else:
            cnt_count += 1

    assert dir_count == 3
    assert cnt_count == 1

    # the last swhid must be a content related to the unknown directory
    # "sample-folder-policy/toexclude"
    assert (
        backend_swhids_requests[5]
        == "swh:1:cnt:5f1cfce26640056bed3710cfaf3062a6a326a119"
    )


def test_directory_priority_policy(
    live_server, aiosession, event_loop, source_tree_policy, tmp_requests
):
    open(tmp_requests, "w").close()
    api_url = url_for("index", _external=True)

    nodes_data = MerkleNodeInfo()
    init_merkle_node_info(source_tree_policy, nodes_data, {"known"})
    policy = DirectoryPriority(source_tree_policy, nodes_data)
    client = Client(api_url, aiosession)
    event_loop.run_until_complete(policy.run(client))

    backend_swhids_requests = get_backend_swhids_order(tmp_requests)

    for swhid in backend_swhids_requests[0:4]:
        assert CoreSWHID.from_string(swhid).object_type == ObjectType.DIRECTORY

    for swhid in backend_swhids_requests[5:]:
        assert CoreSWHID.from_string(swhid).object_type == ObjectType.CONTENT


def test_file_priority_policy(
    live_server, aiosession, event_loop, source_tree_policy, tmp_requests
):
    open(tmp_requests, "w").close()
    api_url = url_for("index", _external=True)

    nodes_data = MerkleNodeInfo()
    init_merkle_node_info(source_tree_policy, nodes_data, {"known"})
    policy = FilePriority(source_tree_policy, nodes_data)
    client = Client(api_url, aiosession)
    event_loop.run_until_complete(policy.run(client))

    backend_swhids_requests = get_backend_swhids_order(tmp_requests)

    for swhid in backend_swhids_requests[0:4]:
        assert CoreSWHID.from_string(swhid).object_type == ObjectType.CONTENT

    for swhid in backend_swhids_requests[5:]:
        assert CoreSWHID.from_string(swhid).object_type == ObjectType.DIRECTORY


def test_greedy_bfs_policy(
    live_server, event_loop, aiosession, big_source_tree, tmp_requests
):
    open(tmp_requests, "w").close()
    api_url = url_for("index", _external=True)

    nodes_data = MerkleNodeInfo()
    init_merkle_node_info(big_source_tree, nodes_data, {"known"})
    policy = GreedyBFS(big_source_tree, nodes_data)
    client = Client(api_url, aiosession)
    event_loop.run_until_complete(policy.run(client))

    backend_swhids_requests = get_backend_swhids_order(tmp_requests)

    last_swhid = backend_swhids_requests[-1]
    assert CoreSWHID.from_string(last_swhid).object_type == ObjectType.CONTENT


def test_randomdir_policy(
    live_server,
    event_loop,
    aiosession,
    big_source_tree,
    tmp_requests,
    tmp_accesses,
    mocker,
):
    # This is harder to test with exact assertions due to the random nature
    # of our sampling algorithm and everything else that can be random.
    # Setting random.seed has failed to produce stable results.
    # TODO figure out why?

    open(tmp_requests, "w").close()
    open(tmp_accesses, "w").close()
    api_url = url_for("index", _external=True)

    nodes_data = MerkleNodeInfo()
    init_merkle_node_info(big_source_tree, nodes_data, {"known"})
    policy = RandomDirSamplingPriority(big_source_tree, nodes_data)
    client = Client(api_url, aiosession)
    event_loop.run_until_complete(policy.run(client))

    backend_swhids_requests = get_backend_swhids_order(tmp_requests)
    # Check that we only query directories in the case where all directories
    # fit in a single request
    assert all(
        CoreSWHID.from_string(swhid).object_type == ObjectType.DIRECTORY
        for swhid in backend_swhids_requests
    )

    assert all(v["known"] is True for k, v in policy.data.items())

    # Check that we only do a single query of 1000 items
    backend_known_requests = get_backend_known_requests(tmp_accesses)
    assert [1000] == backend_known_requests

    # Test with smaller sample sizes to actually trigger the random sampling
    open(tmp_requests, "w").close()
    open(tmp_accesses, "w").close()
    mocker.patch("swh.scanner.policy.discovery.SAMPLE_SIZE", 10)

    nodes_data = MerkleNodeInfo()
    init_merkle_node_info(big_source_tree, nodes_data, {"known"})
    policy = RandomDirSamplingPriority(big_source_tree, nodes_data)
    client = Client(api_url, aiosession)
    event_loop.run_until_complete(policy.run(client))

    assert all(v["known"] is True for k, v in policy.data.items())

    # Check that we only do at least two queries of < 10 items
    backend_known_requests = get_backend_known_requests(tmp_accesses)
    assert len(backend_known_requests) >= 2
    assert all(length <= 10 for length in backend_known_requests)


@pytest.mark.asyncio
async def test_greedy_bfs_get_nodes_chunks(live_server, aiosession, big_source_tree):
    api_url = url_for("index", _external=True)

    nodes_data = MerkleNodeInfo()
    init_merkle_node_info(big_source_tree, nodes_data, {"known"})
    policy = GreedyBFS(big_source_tree, nodes_data)
    client = Client(api_url, aiosession)
    chunks = [
        n_chunk
        async for n_chunk in policy.get_nodes_chunks(
            client, source_size(big_source_tree)
        )
    ]
    assert len(chunks) == 2
    assert chunks[1][-1].object_type == "content"



# Copyright (C) 2020  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

# WARNING: do not import unnecessary things here to keep cli startup time under
# control
import os
from pathlib import Path
import sys
from typing import Any, Dict, Optional

import click
from importlib_metadata import version
import yaml

from swh.core import config
from swh.core.cli import CONTEXT_SETTINGS
from swh.core.cli import swh as swh_cli_group

from .exceptions import DBError

# Config for the "serve" option
BACKEND_DEFAULT_PORT = 5011

# All generic config code should reside in swh.core.config
CONFIG_ENVVAR = "SWH_CONFIG_FILE"
DEFAULT_CONFIG_PATH = os.path.join(click.get_app_dir("swh"), "global.yml")

SWH_API_ROOT = "https://archive.softwareheritage.org/api/1/"
DEFAULT_CONFIG: Dict[str, Any] = {
    "web-api": {
        "url": SWH_API_ROOT,
        "auth-token": None,
    }
}


CONFIG_FILE_HELP = f"""Configuration file:

\b
The CLI option or the environment variable will fail if invalid.
CLI option is checked first.
Then, environment variable {CONFIG_ENVVAR} is checked.
Then, if cannot load the default path, a set of default values are used.
Default config path is {DEFAULT_CONFIG_PATH}.
Default config values are:

\b
{yaml.dump(DEFAULT_CONFIG)}"""
SCANNER_HELP = f"""Software Heritage Scanner tools.

{CONFIG_FILE_HELP}"""


def setup_config(ctx, api_url):
    config = ctx.obj["config"]
    if api_url:
        if not api_url.endswith("/"):
            api_url += "/"
        config["web-api"]["url"] = api_url

    return config


def check_auth(config):
    """check there is some authentication configured

    Issue a warning otherwise"""
    web_api_conf = config["web-api"]
    if web_api_conf["url"] == SWH_API_ROOT and not web_api_conf.get("auth-token"):
        # Only warn for the production API
        #
        # XXX We should probably warn at the time of the creation of the HTTP
        # Client, after checking if the token is actually valid.
        msg = "Warning: you are not authenticated with the Software Heritage API\n"
        msg += "login to get a higher rate-limit"
        click.echo(click.style(msg, fg="red"), file=sys.stderr)
        msg = "See `swh scanner login -h` for more information."
        click.echo(click.style(msg, fg="yellow"), file=sys.stderr)


@swh_cli_group.group(
    name="scanner",
    context_settings=CONTEXT_SETTINGS,
    help=SCANNER_HELP,
)
@click.option(
    "-C",
    "--config-file",
    default=None,
    type=click.Path(exists=False, dir_okay=False, path_type=str),
    help="""YAML configuration file""",
)
@click.version_option(
    version=version("swh.scanner"),
    prog_name="swh.scanner",
)
@click.pass_context
def scanner(ctx, config_file: Optional[str]):

    env_config_path = os.environ.get(CONFIG_ENVVAR)

    # read_raw_config do not fail if file does not exist, so check it beforehand
    # while enforcing loading priority
    if config_file:
        if not config.exists_accessible(config_file):
            raise click.BadParameter(
                f"File '{config_file}' cannot be opened.", param_hint="--config-file"
            )
    elif env_config_path:
        if not config.exists_accessible(env_config_path):
            raise click.BadParameter(
                f"File '{env_config_path}' cannot be opened.", param_hint=CONFIG_ENVVAR
            )
        config_file = env_config_path
    elif config.exists_accessible(DEFAULT_CONFIG_PATH):
        config_file = DEFAULT_CONFIG_PATH

    conf = DEFAULT_CONFIG
    if config_file is not None:
        conf = config.read_raw_config(config_file)
        conf = config.merge_configs(DEFAULT_CONFIG, conf)
    else:
        config_file = DEFAULT_CONFIG_PATH

    ctx.ensure_object(dict)
    ctx.obj["config_path"] = Path(config_file)
    ctx.obj["config"] = conf


@scanner.command(name="login")
@click.option(
    "-f",
    "--force/--no-force",
    default=False,
    help="Proceed even if a token is already present in the config",
)
@click.pass_context
def login(ctx, force):
    """Perform the necessary step to log yourself in the API

    You will need to first create an account before running this operation. To
    create an account, visit: https://archive.softwareheritage.org/
    """
    context = ctx.obj

    # Check we are actually talking to the Software Heritage itself.
    web_api_config = context["config"]["web-api"]
    current_url = web_api_config["url"]
    config_path = context["config_path"]
    if current_url != SWH_API_ROOT:
        msg = "`swh scanner login` only works with the Software Heritage API\n"
        click.echo(click.style(msg, fg="red"), file=sys.stderr)
        msg = f"Configured in '%s' as web-api.url={current_url}\n"
        msg %= click.format_filename(bytes(config_path))
        click.echo(click.style(msg, fg="red"), file=sys.stderr)
        ctx.exit(1)

    # Check for an existing value in the configuration
    if web_api_config.get("auth-token") is not None:
        click.echo(click.style("You appear to already be logged in.", fg="green"))
        if not force:
            click.echo("Hint: use `--force` to overwrite the current token")
            ctx.exit()
        click.echo(click.style("Continuing because of `--force`.", fg="yellow"))

    # Obtain a valid token through the API
    #
    # Coming from the swh auth generate-token code
    # (this command might eventually move there)
    from getpass import getpass

    from swh.auth.keycloak import (
        KeycloakError,
        KeycloakOpenIDConnect,
        keycloak_error_message,
    )

    msg = "Please enter your SWH Archive credentials"
    click.echo(click.style(msg, fg="yellow"))
    msg = "If you do not already have an account, create one one at:"
    click.echo(click.style(msg, fg="yellow"))
    msg = "    https://archive.softwareheritage.org/"
    click.echo(click.style(msg, fg="yellow"))
    username = click.prompt("username")
    password = getpass()
    try:
        url = "https://auth.softwareheritage.org/auth/"
        realm = "SoftwareHeritage"
        client = "swh-web"
        oidc_client = KeycloakOpenIDConnect(url, realm, client)
        scope = "openid offline_access"
        oidc_info = oidc_client.login(username, password, scope)
        token = oidc_info["refresh_token"]
        msg = "token retrieved successfully"
        click.echo(click.style(msg, fg="green"))
    except KeycloakError as ke:
        print(keycloak_error_message(ke))
        click.exit(1)

    # Write the new token into the file.
    web_api_config["auth-token"] = token
    # TODO use ruamel.yaml to preserve comments in config file
    config_path.parent.mkdir(parents=True, exist_ok=True)
    config_path.write_text(yaml.safe_dump(context["config"]))
    msg = "\nConfiguration file '%s' written successfully."
    msg %= click.format_filename(bytes(config_path))
    click.echo(click.style(msg, fg="green"))
    click.echo("`swh scanner` will now be authenticated with the new token.")


@scanner.command(name="scan")
@click.argument("root_path", default=".", type=click.Path(exists=True))
@click.option(
    "-u",
    "--api-url",
    default=None,
    metavar="API_URL",
    show_default=True,
    help="URL for the api request",
)
@click.option(
    "--exclude",
    "-x",
    "patterns",
    metavar="PATTERN",
    multiple=True,
    help="Exclude directories using glob patterns \
    (e.g., ``*.git`` to exclude all .git directories)",
)
@click.option(
    "-f",
    "--output-format",
    "out_fmt",
    default="summary",
    show_default=True,
    type=click.Choice(
        ["summary", "text", "json", "ndjson", "sunburst"], case_sensitive=False
    ),
    help="The output format",
)
@click.option(
    "-i", "--interactive", is_flag=True, help="Show the result in a dashboard"
)
@click.option(
    "-p",
    "--policy",
    default="auto",
    show_default=True,
    type=click.Choice(
        ["auto", "bfs", "greedybfs", "filepriority", "dirpriority", "randomdir"]
    ),
    help="The scan policy.",
)
@click.option(
    "-e",
    "--extra-info",
    "extra_info",
    multiple=True,
    type=click.Choice(["origin"]),
    help="Add selected additional information about known software artifacts.",
)
@click.pass_context
def scan(ctx, root_path, api_url, patterns, out_fmt, interactive, policy, extra_info):
    """Scan a source code project to discover files and directories already
    present in the archive.

    The command can provide different output using the --output-format option:\n
    \b
      summary: display a general summary of what the scanner found

      text: display the scan result as a text based tree-like view of all the
            file, using color to indicate the file status.

      json: write all collected data on standard output as JSON

      json: write all collected data on standard output as Newline Delimited JSON

      sunburst: produce a dynamic chart as .html file. (in $PWD/chart.html)

    The source code project can be checked using different policies that can be set
    using the -p/--policy option:\n
    \b
      auto: it selects the best policy based on the source code, for codebase(s)
      with less than 1000 file/dir contents all the nodes will be queried.

      bfs: scan the source code in the BFS order, checking unknown directories only.

    \b
      greedybfs: same as "bfs" policy, but lookup the status of source code artifacts
      in chunks, in order to minimize the number of Web API round-trips with the
      archive.

    \b
      filepriority: scan all the source code file contents, checking only unset
      directories. (useful if the codebase contains a lot of source files)

      dirpriority: scan all the source code directories and check only unknown
      directory contents.

      randomdir: scan the source code using a random Merkle search on directories.

    Other information about software artifacts could be specified with the -e/
    --extra-info option:\n
    \b
      origin: search the origin url of each source code files/dirs using the in-memory
      compressed graph."""
    import swh.scanner.scanner as scanner

    config = setup_config(ctx, api_url)
    check_auth(config)
    extra_info = set(extra_info)
    scanner.scan(config, root_path, patterns, out_fmt, interactive, policy, extra_info)


@scanner.group("db", help="Manage local knowledge base for swh-scanner")
@click.pass_context
def db(ctx):
    pass


@db.command("import")
@click.option(
    "-i",
    "--input",
    "input_file",
    metavar="INPUT_FILE",
    required=True,
    type=click.File("r"),
    help="A file containing SWHIDs",
)
@click.option(
    "-o",
    "--output",
    "output_file_db",
    metavar="OUTPUT_DB_FILE",
    required=True,
    show_default=True,
    help="The name of the generated sqlite database",
)
@click.option(
    "-s",
    "--chunk-size",
    "chunk_size",
    default="10000",
    metavar="SIZE",
    show_default=True,
    type=int,
    help="The chunk size ",
)
@click.pass_context
def import_(ctx, chunk_size, input_file, output_file_db):
    """Create SQLite database of known SWHIDs from a textual list of SWHIDs"""
    from .db import Db

    db = Db(output_file_db)
    cur = db.conn.cursor()
    try:
        db.create_from(input_file, chunk_size, cur)
        db.close()
    except DBError as e:
        ctx.fail("Failed to import SWHIDs into database: {0}".format(e))


@db.command("serve")
@click.option(
    "-h",
    "--host",
    metavar="HOST",
    default="127.0.0.1",
    show_default=True,
    help="The host of the API server",
)
@click.option(
    "-p",
    "--port",
    metavar="PORT",
    default=f"{BACKEND_DEFAULT_PORT}",
    show_default=True,
    help="The port of the API server",
)
@click.option(
    "-f",
    "--db-file",
    "db_file",
    metavar="DB_FILE",
    default="SWHID_DB.sqlite",
    show_default=True,
    type=click.Path(exists=True),
    help="An sqlite database file (it can be generated with: 'swh scanner db import')",
)
@click.pass_context
def serve(ctx, host, port, db_file):
    """Start an API service using the sqlite database generated with the "db import"
    option."""
    import swh.scanner.backend as backend

    from .db import Db

    db = Db(db_file)
    backend.run(host, port, db)
    db.close()


def main():
    return scanner(auto_envvar_prefix="SWH_SCANNER")


if __name__ == "__main__":
    main()



# Copyright (C) 2020-2021 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import asyncio
from typing import Any, Dict, Iterable

import aiohttp

from swh.model.cli import model_of_dir
from swh.model.from_disk import Directory

from .client import Client
from .data import MerkleNodeInfo, add_origin, init_merkle_node_info
from .output import Output
from .policy import (
    QUERY_LIMIT,
    DirectoryPriority,
    FilePriority,
    GreedyBFS,
    LazyBFS,
    QueryAll,
    RandomDirSamplingPriority,
    source_size,
)


async def run(
    config: Dict[str, Any],
    policy,
    source_tree: Directory,
    nodes_data: MerkleNodeInfo,
    extra_info: set,
) -> None:
    """Scan a given source code according to the policy given in input."""
    api_url = config["web-api"]["url"]

    if config["web-api"]["auth-token"]:
        headers = {"Authorization": f"Bearer {config['web-api']['auth-token']}"}
    else:
        headers = {}

    async with aiohttp.ClientSession(headers=headers, trust_env=True) as session:
        client = Client(api_url, session)
        for info in extra_info:
            if info == "known":
                await policy.run(client)
            elif info == "origin":
                await add_origin(source_tree, nodes_data, client)
            else:
                raise Exception(f"The information '{info}' cannot be retrieved")


def get_policy_obj(source_tree: Directory, nodes_data: MerkleNodeInfo, policy: str):
    if policy == "auto":
        return (
            QueryAll(source_tree, nodes_data)
            if source_size(source_tree) <= QUERY_LIMIT
            else LazyBFS(source_tree, nodes_data)
        )
    elif policy == "bfs":
        return LazyBFS(source_tree, nodes_data)
    elif policy == "greedybfs":
        return GreedyBFS(source_tree, nodes_data)
    elif policy == "filepriority":
        return FilePriority(source_tree, nodes_data)
    elif policy == "dirpriority":
        return DirectoryPriority(source_tree, nodes_data)
    elif policy == "randomdir":
        return RandomDirSamplingPriority(source_tree, nodes_data)
    else:
        raise Exception(f"policy '{policy}' not found")


# here is a set of directory we should disregard
#
# TODO: make its usage configurable
# TODO: make it extensible through configuration
COMMON_EXCLUDE_PATTERNS = [
    b".bzr",
    b".coverage",
    b"*.egg-info",
    b".eggs",
    b".git",
    b".hg",
    b".mypy_cache",
    b"__pycache__",
    b".svn",
    b".tox",
]
COMMON_EXCLUDE_PATTERNS.extend([b"*/" + p for p in COMMON_EXCLUDE_PATTERNS])


def scan(
    config: Dict[str, Any],
    root_path: str,
    exclude_patterns: Iterable[str],
    out_fmt: str,
    interactive: bool,
    policy: str,
    extra_info: set,
):
    """Scan a source code project to discover files and directories already
    present in the archive"""
    converted_patterns = [pattern.encode() for pattern in exclude_patterns]
    converted_patterns.extend(COMMON_EXCLUDE_PATTERNS)
    source_tree = model_of_dir(root_path.encode(), converted_patterns)

    nodes_data = MerkleNodeInfo()
    extra_info.add("known")
    init_merkle_node_info(source_tree, nodes_data, extra_info)

    policy = get_policy_obj(source_tree, nodes_data, policy)

    loop = asyncio.get_event_loop()
    loop.run_until_complete(run(config, policy, source_tree, nodes_data, extra_info))

    out = Output(root_path, nodes_data, source_tree)
    if interactive:
        out.show("interactive")
    else:
        out.show(out_fmt)



# Copyright (C) 2020-2021 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from flask import url_for
import pytest

from swh.scanner.data import MerkleNodeInfo, init_merkle_node_info
from swh.scanner.policy import DirectoryPriority, FilePriority, LazyBFS, QueryAll
from swh.scanner.scanner import get_policy_obj, run

from .data import unknown_swhids


@pytest.mark.options(debug=False)
def test_app(app):
    assert not app.debug


def test_get_policy_obj_auto(source_tree, nodes_data):
    assert isinstance(get_policy_obj(source_tree, nodes_data, "auto"), QueryAll)


def test_get_policy_obj_bfs(big_source_tree, nodes_data):
    # check that the policy object is the LazyBFS if the source tree contains more than
    # 1000 nodes
    assert isinstance(get_policy_obj(big_source_tree, nodes_data, "auto"), LazyBFS)


def test_scanner_result_bfs(live_server, event_loop, source_tree):
    api_url = url_for("index", _external=True)
    config = {"web-api": {"url": api_url, "auth-token": None}}

    nodes_data = MerkleNodeInfo()
    init_merkle_node_info(source_tree, nodes_data, {"known"})
    policy = LazyBFS(source_tree, nodes_data)
    event_loop.run_until_complete(
        run(config, policy, source_tree, nodes_data, {"known"})
    )
    for node in source_tree.iter_tree():
        if str(node.swhid()) in unknown_swhids:
            assert nodes_data[node.swhid()]["known"] is False
        else:
            assert nodes_data[node.swhid()]["known"] is True


def test_scanner_result_file_priority(live_server, event_loop, source_tree):
    api_url = url_for("index", _external=True)
    config = {"web-api": {"url": api_url, "auth-token": None}}

    nodes_data = MerkleNodeInfo()
    init_merkle_node_info(source_tree, nodes_data, {"known"})
    policy = FilePriority(source_tree, nodes_data)
    event_loop.run_until_complete(
        run(config, policy, source_tree, nodes_data, {"known"})
    )
    for node in source_tree.iter_tree():
        if str(node.swhid()) in unknown_swhids:
            assert nodes_data[node.swhid()]["known"] is False
        else:
            assert nodes_data[node.swhid()]["known"] is True


def test_scanner_result_directory_priority(live_server, event_loop, source_tree):
    api_url = url_for("index", _external=True)
    config = {"web-api": {"url": api_url, "auth-token": None}}

    nodes_data = MerkleNodeInfo()
    init_merkle_node_info(source_tree, nodes_data, {"known"})
    policy = DirectoryPriority(source_tree, nodes_data)
    event_loop.run_until_complete(
        run(config, policy, source_tree, nodes_data, {"known"})
    )
    for node in source_tree.iter_tree():
        if str(node.swhid()) in unknown_swhids:
            assert nodes_data[node.swhid()]["known"] is False
        else:
            assert nodes_data[node.swhid()]["known"] is True


def test_scanner_result_query_all(live_server, event_loop, source_tree):
    api_url = url_for("index", _external=True)
    config = {"web-api": {"url": api_url, "auth-token": None}}

    nodes_data = MerkleNodeInfo()
    init_merkle_node_info(source_tree, nodes_data, {"known"})
    policy = QueryAll(source_tree, nodes_data)
    event_loop.run_until_complete(
        run(config, policy, source_tree, nodes_data, {"known"})
    )
    for node in source_tree.iter_tree():
        if str(node.swhid()) in unknown_swhids:
            assert nodes_data[node.swhid()]["known"] is False
        else:
            assert nodes_data[node.swhid()]["known"] is True



# Copyright (C) 2021 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from pathlib import Path

from swh.scanner.data import get_directory_data
from swh.scanner.plot import (
    build_hierarchical_df,
    compute_max_depth,
    generate_df_from_dirs,
)


def test_max_depth(source_tree, source_tree_dirs):
    dirs = [Path(dir_path) for dir_path in source_tree_dirs]
    max_depth = compute_max_depth(dirs)
    assert max_depth == 2


def test_generate_df_from_dirs(source_tree, source_tree_dirs, nodes_data):
    root = Path(source_tree.data["path"].decode())
    dirs = [Path(dir_path) for dir_path in source_tree_dirs]
    dirs_data = get_directory_data(root, source_tree, nodes_data)
    max_depth = compute_max_depth(dirs)
    metrics_columns = ["contents", "known"]
    levels_columns = ["lev" + str(i) for i in range(max_depth)]
    df_columns = levels_columns + metrics_columns

    actual_df = generate_df_from_dirs(dirs_data, df_columns, max_depth)

    expected_lev0_path = ["bar", "foo", "toexclude"]
    expected_lev1_path = ["bar/barfoo", "bar/barfoo2"]

    df_lev0 = actual_df["lev0"].tolist()
    df_lev1 = actual_df["lev1"].tolist()

    for path in expected_lev0_path:
        assert path in df_lev0

    for path in expected_lev1_path:
        assert path in df_lev1

    assert actual_df["contents"].sum() == 6
    assert actual_df["known"].sum() == 6


def test_build_hierarchical_df(source_tree, source_tree_dirs, nodes_data):
    root = Path(source_tree.data["path"].decode())
    dirs = [Path(dir_path) for dir_path in source_tree_dirs]
    dirs_data = get_directory_data(root, source_tree, nodes_data)
    max_depth = compute_max_depth(dirs)
    metrics_columns = ["contents", "known"]
    levels_columns = ["lev" + str(i) for i in range(max_depth)]
    df_columns = levels_columns + metrics_columns

    actual_df = generate_df_from_dirs(dirs_data, df_columns, max_depth)

    actual_result = build_hierarchical_df(
        actual_df, levels_columns, metrics_columns, root
    )

    assert actual_result["parent"][0] == "bar"
    assert actual_result["parent"][1] == "foo"
    assert actual_result["contents"][1] == 3
    assert actual_result["id"][8] == root
    assert actual_result["known"][8] == 100
